# Copyright (c) 2019-2024 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from http import HTTPStatus

from django.test import TestCase
from django.urls import reverse
from parameterized import parameterized


class ServeUsingFindersTest(TestCase):
    @parameterized.expand(
        [
            # Our app, some arbitrary asset
            ("jawanndenn", "js/html.js", "DENY"),
            # Our app, asset used in <iframe>
            ("jawanndenn", "3rdparty/github-buttons-4.0.1/docs/github-btn.html", "sameorigin"),
            # Arbitrary asset of arbitary other app
            ("django.contrib.admin", "admin/css/responsive.css", "DENY"),
        ]
    )
    def test(self, _app, path, expected_x_frame_options):
        url = reverse("static", kwargs={"path": path})

        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(response["X-Frame-Options"], expected_x_frame_options)



# Generated by Django 3.1 on 2020-08-15 23:45

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("jawanndenn", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="ballot",
            options={"get_latest_by": "modified"},
        ),
        migrations.AlterModelOptions(
            name="poll",
            options={"get_latest_by": "modified"},
        ),
    ]



"""
Django settings for jawanndenn project.

Generated by 'django-admin startproject' using Django 2.2.7.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import copy
import os

from django.conf.global_settings import CACHES
from django.utils.log import DEFAULT_LOGGING

from jawanndenn import DEFAULT_MAX_POLLS, DEFAULT_MAX_VOTES_PER_POLL

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ["JAWANNDENN_SECRET_KEY"]

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get("JAWANNDENN_DEBUG", "False") == "True"

_ALLOWED_HOSTS_DEFAULT = ",".join(
    [
        "127.0.0.1",
        "0.0.0.0",
        "localhost",
    ]
)
ALLOWED_HOSTS = (os.environ.get("JAWANNDENN_ALLOWED_HOSTS") or _ALLOWED_HOSTS_DEFAULT).split(",")
CSRF_TRUSTED_ORIGINS = {
    f"{scheme}://{host}" for scheme in ("http", "https") for host in ALLOWED_HOSTS
}

SECURE_REFERRER_POLICY = "origin"

_USE_POSTGRES = "JAWANNDENN_SQLITE_FILE" not in os.environ
_USE_REDIS_CACHE = "JAWANNDENN_REDIS_HOST" in os.environ

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_extensions",  # for management command "show_urls"
    "jawanndenn",
]

MIDDLEWARE = [
    "jawanndenn.middleware.set_remote_addr_to_x_forwarded_for",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "jawanndenn.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            os.path.join(BASE_DIR, "jawanndenn", "static"),
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "jawanndenn.wsgi.application"

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

if _USE_POSTGRES:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "NAME": os.environ["JAWANNDENN_POSTGRES_NAME"],
            "USER": os.environ["JAWANNDENN_POSTGRES_USER"],
            "PASSWORD": os.environ["JAWANNDENN_POSTGRES_PASSWORD"],
            "HOST": os.environ["JAWANNDENN_POSTGRES_HOST"],
            "PORT": os.environ["JAWANNDENN_POSTGRES_PORT"],
        }
    }
else:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": os.environ["JAWANNDENN_SQLITE_FILE"],
        }
    }

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# Logging
# https://docs.djangoproject.com/en/2.2/ref/settings/#std:setting-LOGGING

LOGGING = copy.copy(DEFAULT_LOGGING)
LOGGING["handlers"]["console"]["filters"].remove("require_debug_true")
if not DEBUG:
    LOGGING["handlers"]["console"]["level"] = "ERROR"

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",  # noqa: E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",  # noqa: E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",  # noqa: E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",  # noqa: E501
    },
]

# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Settings specific to jawanndenn

JAWANNDENN_URL_PREFIX = os.environ.get("JAWANNDENN_URL_PREFIX", "").strip("/")

if JAWANNDENN_URL_PREFIX != "":
    JAWANNDENN_URL_PREFIX = "/" + JAWANNDENN_URL_PREFIX

JAWANNDENN_MAX_POLLS = int(os.environ.get("JAWANNDENN_MAX_POLLS", str(DEFAULT_MAX_POLLS)))
JAWANNDENN_MAX_VOTES_PER_POLL = int(
    os.environ.get("JAWANNDENN_MAX_VOTES_PER_POLL", str(DEFAULT_MAX_VOTES_PER_POLL))
)

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = JAWANNDENN_URL_PREFIX + "/static/"

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, "jawanndenn", "static"),
]

# Rate limiting

if _USE_REDIS_CACHE:
    RATELIMIT_USE_CACHE = "ratelimit"

    _REDIS_URL = "redis://{}:{}/1".format(
        os.environ["JAWANNDENN_REDIS_HOST"], os.environ["JAWANNDENN_REDIS_PORT"]
    )

    CACHES = copy.copy(CACHES)
    CACHES[RATELIMIT_USE_CACHE] = {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": _REDIS_URL,
        "OPTIONS": {
            "SERIALIZER": "django_redis.serializers.msgpack.MSGPackSerializer",
        },
    }

# Sentry integration

_SENTRY_DSN = os.environ.get("JAWANNDENN_SENTRY_DSN")

if _SENTRY_DSN:
    import sentry_sdk
    from sentry_sdk.integrations.django import DjangoIntegration

    _sentry_integrations = [DjangoIntegration()]

    if _USE_REDIS_CACHE:
        from sentry_sdk.integrations.redis import RedisIntegration

        _sentry_integrations.append(RedisIntegration())

    sentry_sdk.init(
        dsn=_SENTRY_DSN,
        integrations=_sentry_integrations,
        send_default_pii=False,
        traces_sample_rate=1.0,
    )



# Copyright (C) 2020 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from django.test import RequestFactory, TestCase
from parameterized import parameterized

from jawanndenn.middleware import set_remote_addr_to_x_forwarded_for


class SetRemoteAddrToXForwardedForMiddlewareTest(TestCase):
    @parameterized.expand(
        [
            (None, "127.0.0.1"),  # inserted by Django
            ("2.2.2.2", "2.2.2.2"),
            ("  2.2.2.2  ", "2.2.2.2"),
            ("1.1.1.1,2.2.2.2", "1.1.1.1"),
            ("1.1.1.1, 2.2.2.2", "1.1.1.1"),
            ("1.1.1.1 ,2.2.2.2", "1.1.1.1"),
            ("1.1.1.1 , 2.2.2.2", "1.1.1.1"),
        ]
    )
    def test(self, http_x_forwarded_for, expected_remote_addr):
        wrapped_get_response = set_remote_addr_to_x_forwarded_for(
            get_response=lambda request: None
        )
        headers = {}
        if http_x_forwarded_for is not None:
            headers["HTTP_X_FORWARDED_FOR"] = http_x_forwarded_for
        request = RequestFactory().get("/", **headers)

        wrapped_get_response(request)

        self.assertEqual(request.META.get("REMOTE_ADDR"), expected_remote_addr)



# Copyright (c) 2019-2024 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import re

from django.conf import settings
from django.contrib.staticfiles.views import serve
from django.urls import re_path


def _serve_with_headers_fixed(request, path, insecure=False, **kwargs):
    response = serve(request, path, insecure=insecure, **kwargs)

    # Allow loading of github-btn.html in an <iframe>
    if path.startswith("3rdparty/github-buttons-") and path.endswith("/docs/github-btn.html"):
        response["X-Frame-Options"] = "sameorigin"

    return response


def staticfiles_urlpatterns(prefix=None, name="static"):
    """
    Fork of django.contrib.staticfiles.urls.staticfiles_urlpatterns
    that supports DEBUG=False, directory listings, and registering
    a name for the view.

    Also, it uses view _serve_with_headers_fixed above rather than stock
    django.contrib.staticfiles.views.serve to serve files.
    """
    if prefix is None:
        prefix = settings.STATIC_URL
    return [
        re_path(
            r"^%s(?P<path>.*)$" % re.escape(prefix.lstrip("/")),
            _serve_with_headers_fixed,
            kwargs={
                "insecure": not settings.DEBUG,
                "show_indexes": settings.DEBUG,
            },
            name=name,
        ),
    ]



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from http import HTTPStatus
from itertools import chain

from django.conf import settings
from django.contrib import admin
from django.http import HttpResponse
from django.urls import include, path
from django.views.defaults import permission_denied
from django_ratelimit.decorators import ratelimit
from django_ratelimit.exceptions import Ratelimited

from .views.polls import (
    index_get_view,
    poll_data_get_view,
    poll_get_view,
    poll_post_view,
    vote_post_view,
)
from .views.security_txt import security_txt_urlpatterns
from .views.static_files import staticfiles_urlpatterns


class _HttpResponseTooManyRequests(HttpResponse):
    status_code = HTTPStatus.TOO_MANY_REQUESTS

    def __init__(self):
        super().__init__(
            f"<h1>{self.status_code} Too Many Requests</h1>", content_type="text/html"
        )


def _permission_denied_or_too_many_requests(request, exception=None):
    if isinstance(exception, Ratelimited):
        return _HttpResponseTooManyRequests()

    return permission_denied(request, exception)


def _decorate_view_of_url_pattern(decorator, url_pattern):
    url_pattern.callback = decorator(url_pattern.callback)
    return url_pattern


def _decorate_view_triple(decorator, view):
    urlconf_module, _app_name, _namespace = view

    for url_pattern in urlconf_module:
        _decorate_view_of_url_pattern(decorator, url_pattern)

    return view


_limit_read_access = ratelimit(key="user_or_ip", rate="180/m")

_limit_write_access = ratelimit(key="user_or_ip", rate="30/m")

_app_urlpatterns = [
    path("", _limit_read_access(index_get_view), name="frontpage"),
    path("create", _limit_write_access(poll_post_view), name="poll-creation"),
    path("data/<poll_id>", _limit_read_access(poll_data_get_view), name="poll-data"),
    path("poll/<poll_id>", _limit_read_access(poll_get_view), name="poll-detail"),
    path("vote/<poll_id>", _limit_write_access(vote_post_view), name="vote"),
    path("admin/", _decorate_view_triple(_limit_write_access, admin.site.urls)),
]

_url_prefix = getattr(settings, "JAWANNDENN_URL_PREFIX", None)
if _url_prefix:
    from django.views.generic.base import RedirectView

    urlpatterns = [
        path("", _limit_read_access(RedirectView.as_view(url=_url_prefix + "/"))),
        path(_url_prefix.strip("/") + "/", include(_app_urlpatterns)),
    ]
else:
    urlpatterns = _app_urlpatterns

urlpatterns += [
    _decorate_view_of_url_pattern(_limit_read_access, url_pattern)
    for url_pattern in chain(staticfiles_urlpatterns(), security_txt_urlpatterns())
]

handler403 = _permission_denied_or_too_many_requests



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import hashlib
import os

from django.db import models
from django.urls import reverse
from django.utils.timezone import now
from django_extensions.db.models import TimeStampedModel


def _get_random_sha256():
    return hashlib.sha256(os.urandom(256 // 8)).hexdigest()


class PollQuerySet(models.QuerySet):
    def expired(self):
        return self.filter(expires_at__lt=now())


class Poll(TimeStampedModel):
    slug = models.CharField(max_length=64, default=_get_random_sha256, unique=True)
    title = models.CharField(max_length=255)
    equal_width = models.BooleanField(default=False)
    expires_at = models.DateTimeField(null=True)

    objects = PollQuerySet.as_manager()

    def get_absolute_url(self):
        return reverse("poll-detail", args=[self.slug])


class PollOption(models.Model):
    poll = models.ForeignKey(Poll, on_delete=models.CASCADE, related_name="options")
    position = models.PositiveSmallIntegerField()  # starting at 0
    name = models.CharField(max_length=255)

    class Meta:
        unique_together = ("poll", "position")


class Ballot(TimeStampedModel):
    poll = models.ForeignKey(Poll, on_delete=models.CASCADE, related_name="ballots")
    voter_name = models.CharField(max_length=255)


class Vote(models.Model):
    ballot = models.ForeignKey(Ballot, related_name="votes", on_delete=models.CASCADE)
    option = models.ForeignKey(PollOption, on_delete=models.CASCADE)
    yes = models.BooleanField()

    class Meta:
        unique_together = ("ballot", "option")



# Generated by Django 3.1 on 2020-08-15 23:54

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("jawanndenn", "0002_django_extensions_3_0_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="poll",
            name="expires_at",
            field=models.DateTimeField(null=True),
        ),
    ]



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from dateutil.relativedelta import relativedelta
from django.db import transaction
from django.utils import timezone
from rest_framework.fields import BooleanField, CharField, ChoiceField, ListField
from rest_framework.serializers import Serializer

from jawanndenn.markup import safe_html
from jawanndenn.models import Poll, PollOption


class _PollLifetime:
    MONTH = "month"
    WEEK = "week"
    CHOICES = (MONTH, WEEK)

    @classmethod
    def to_relativedelta(cls, lifetime):
        return relativedelta(
            **(
                {
                    cls.MONTH: {"months": 1},
                    cls.WEEK: {"days": 7},
                }[lifetime]
            )
        )


class PollConfigSerializer(Serializer):
    equal_width = BooleanField(default=False)
    title = CharField()
    options = ListField(child=CharField(), allow_empty=False)
    lifetime = ChoiceField(choices=_PollLifetime.CHOICES, default=_PollLifetime.MONTH)

    def create(self, validated_data):
        poll_expires_at = timezone.now() + _PollLifetime.to_relativedelta(
            validated_data["lifetime"]
        )
        poll_equal_width = validated_data["equal_width"]
        poll_title = safe_html(validated_data["title"])
        poll_option_names = [
            safe_html(str(option_name)) for option_name in validated_data["options"]
        ]

        with transaction.atomic():
            poll = Poll.objects.create(
                title=poll_title, expires_at=poll_expires_at, equal_width=poll_equal_width
            )
            for i, option_name in enumerate(poll_option_names):
                PollOption.objects.create(poll=poll, position=i, name=option_name)

            return poll



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import datetime
from http import HTTPStatus
from unittest.mock import patch

import rapidjson as json
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone
from django.utils.timezone import now
from parameterized import parameterized

from jawanndenn.models import Ballot, Poll
from jawanndenn.tests.factories import BallotFactory, PollFactory, PollOptionFactory, VoteFactory
from jawanndenn.tests.helpers import RELEASE_SAVEPOINT, SAVEPOINT, SELECT


class AdminLoginPageTest(TestCase):
    def test(self):
        url = reverse("admin:index")
        response = self.client.get(url, follow=True)
        self.assertContains(response, "Django administration")


class IndexGetViewTest(TestCase):
    def test(self):
        url = reverse("frontpage")

        response = self.client.get(url)

        self.assertContains(response, "Create a new poll")


class PollPostViewTest(TestCase):
    url = reverse("poll-creation")

    def test_malformed__config_not_json(self):
        data = {
            "config": "not JSON",
        }
        before_creation = timezone.now()

        response = self.client.post(self.url, data)

        self.assertEqual(response.status_code, HTTPStatus.BAD_REQUEST)
        self.assertFalse(Poll.objects.filter(created__gte=before_creation).exists())

    @parameterized.expand(
        [
            ("month", datetime.date(2020, 9, 16)),
            ("week", datetime.date(2020, 8, 23)),
        ]
    )
    def test_lifetime_sets_expires_at_properly(self, lifetime, expected_expiry_date):
        created_at = now().replace(2020, 8, 16)
        data = {
            "config": json.dumps(
                {
                    "lifetime": lifetime,
                    "options": ["option1", "option2"],
                    "title": "title1",
                }
            ),
        }

        with patch.object(timezone, "now", return_value=created_at):
            response = self.client.post(self.url, data)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        poll = Poll.objects.get(created=created_at)
        self.assertEqual(response.url, poll.get_absolute_url())
        self.assertEqual(poll.expires_at.date(), expected_expiry_date)

    def test_well_formed(self):
        poll_title = "Some short title"
        poll_option_names = ["Option One", "Option Two"]
        poll_equal_width = True
        data = {
            "config": json.dumps(
                {
                    "equal_width": poll_equal_width,
                    "title": poll_title,
                    "options": poll_option_names,
                }
            ),
        }
        before_creation = timezone.now()

        response = self.client.post(self.url, data)

        poll = Poll.objects.get(created__gte=before_creation)
        self.assertEqual(poll.equal_width, poll_equal_width)
        self.assertEqual(
            list(poll.options.order_by("position").values_list("name", flat=True)),
            poll_option_names,
        )
        self.assertEqual(poll.title, poll_title)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, poll.get_absolute_url())


class PollDataGetViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.poll = PollFactory(equal_width=True)
        cls.poll_options = [PollOptionFactory(poll=cls.poll) for _ in range(3)]
        cls.votes_of_ballot = {}

        for user_index in range(2):
            ballot = BallotFactory(poll=cls.poll)
            for option_index, option in enumerate(cls.poll_options):
                vote = VoteFactory(ballot=ballot, option=option, yes=(option_index == user_index))
                cls.votes_of_ballot.setdefault(ballot, []).append(vote)

    def test_poll_exists(self):
        url = reverse("poll-data", args=[self.poll.slug])
        expected_data = {
            "config": {
                "equal_width": self.poll.equal_width,
                "options": [option.name for option in self.poll_options],
                "title": self.poll.title,
            },
            "votes": [
                [ballot.voter_name, [v.yes for v in votes]]
                for ballot, votes in self.votes_of_ballot.items()
            ],
        }

        with self.assertNumQueries(
            SAVEPOINT
            + SELECT("poll")
            + SELECT("poll option names")
            + SELECT("votes + ballots")
            + RELEASE_SAVEPOINT
        ):
            response = self.client.get(url)

        actual_data = json.loads(response.content)
        self.assertEqual(actual_data, expected_data)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_poll_does_not_exist(self):
        url = reverse("poll-data", args=["whatever"])

        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)


class PollGetViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.poll = PollFactory()

    def test_poll_exists(self):
        url = self.poll.get_absolute_url()

        response = self.client.get(url)

        self.assertContains(response, "Vote!")

    def test_poll_does_not_exist(self):
        url = reverse("poll-detail", args=["no_such_poll"])

        response = self.client.get(url)

        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)


class VotePostViewTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.poll = PollFactory()
        for _ in range(2):
            PollOptionFactory(poll=cls.poll)

    def test_poll_exists(self):
        url = reverse("vote", args=[self.poll.slug])
        voter_name = "Maria"
        data = {
            "voterName": voter_name,
            "option0": "on",
            "option1": "off",
        }

        response = self.client.post(url, data)

        ballot = Ballot.objects.get(poll=self.poll)
        self.assertEqual(
            list(ballot.votes.order_by("option__position").values_list("yes", flat=True)),
            [True, False],
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, self.poll.get_absolute_url())

    def test_poll_does_not_exist(self):
        url = reverse("vote", args=["no_such_poll"])
        data = {}

        response = self.client.post(url, data)

        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import argparse
import logging
import os
import signal
import string
import subprocess
import sys
from unittest.mock import patch

from jawanndenn import DEFAULT_MAX_POLLS, DEFAULT_MAX_VOTES_PER_POLL
from jawanndenn.metadata import VERSION_STR

_log = logging.getLogger(__name__)


def _require_hash_randomization():
    if not sys.flags.hash_randomization:
        _log.info("Hash randomization found to be disabled.")
        if os.environ.get("PYTHONHASHSEED") == "random":
            _log.error("Unexpected re-execution loop detected, shutting down.")
            sys.exit(1)

        _log.info("Re-executing with hash randomization enabled...")
        env = os.environ.copy()
        env["PYTHONHASHSEED"] = "random"
        argv = [sys.executable] + sys.argv
        os.execve(argv[0], argv, env)


def _generate_random_printable_django_secret_key():
    alphabet = "".join(
        b for b in string.printable if b not in string.whitespace and b not in "'\\"
    )
    # reduce number of retries while not introducing modulo bias
    alphabet = "".join(alphabet for _ in range(256 // len(alphabet)))
    chars = []
    while len(chars) < 50:
        index = ord(os.urandom(1))
        if index >= len(alphabet):  # detect and avoid modulo bias
            continue
        chars.append(alphabet[index])
    return "".join(chars)


def _process_django_secret_key_file(filename):
    secret_key_encoding = "utf-8"
    try:
        with open(filename, encoding=secret_key_encoding) as f:
            secret_key = f.read()
    except FileNotFoundError:
        secret_key = _generate_random_printable_django_secret_key()
        with open(filename, "w", encoding=secret_key_encoding) as f:
            f.write(secret_key)
    return secret_key


def _inner_main():
    parser = argparse.ArgumentParser(prog="jawanndenn")
    parser.add_argument("--version", action="version", version=f"%(prog)s {VERSION_STR}")
    parser.add_argument(
        "--debug", action="store_true", default=False, help="Enable debug mode (default: disabled)"
    )
    parser.add_argument(
        "--host",
        default="127.0.0.1",
        metavar="HOST",
        help="Hostname or IP address to listen at (default: %(default)s)",
    )
    parser.add_argument(
        "--port",
        default=8080,
        type=int,
        metavar="PORT",
        help="Port to listen at (default: %(default)s)",
    )
    parser.add_argument(
        "--url-prefix",
        default="",
        metavar="PATH",
        help='Path to prepend to URLs (default: "%(default)s")',
    )
    parser.add_argument(
        "--database-sqlite3",
        default="~/jawanndenn.sqlite3",
        metavar="FILE",
        help="File to write the database to (default: %(default)s)",
    )
    parser.add_argument(
        "--django-secret-key-file",
        default="~/jawanndenn.secret_key",
        metavar="FILE",
        help="File to use for Django secret key data (default: %(default)s)",
    )

    limits = parser.add_argument_group("limit configuration")
    limits.add_argument(
        "--max-polls",
        type=int,
        metavar="COUNT",
        default=DEFAULT_MAX_POLLS,
        help="Maximum number of polls total (default: %(default)s)",
    )
    limits.add_argument(
        "--max-votes-per-poll",
        type=int,
        metavar="COUNT",
        default=DEFAULT_MAX_VOTES_PER_POLL,
        help="Maximum number of votes per poll (default: %(default)s)",
    )

    export_args = parser.add_argument_group("data import/export arguments")
    export_args.add_argument(
        "--dumpdata",
        action="store_true",
        help="Dump a JSON export of the database to standard output, then quit.",
    )
    export_args.add_argument(
        "--loaddata",
        metavar="FILE.json",
        help="Load a JSON export of the database from FILE.json, then quit.",
    )

    options = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if options.debug else logging.INFO)

    if not options.dumpdata and not options.loaddata:
        _require_hash_randomization()

    secret_key = _process_django_secret_key_file(
        os.path.expanduser(options.django_secret_key_file)
    )

    # NOTE: These are read by the the Django settings module
    os.environ["JAWANNDENN_ALLOWED_HOSTS"] = ",".join(
        [options.host, "127.0.0.1", "0.0.0.0", "localhost"]
    )
    os.environ["JAWANNDENN_DEBUG"] = str(options.debug)
    os.environ["JAWANNDENN_MAX_POLLS"] = str(options.max_polls)
    os.environ["JAWANNDENN_MAX_VOTES_PER_POLL"] = str(options.max_votes_per_poll)
    os.environ["JAWANNDENN_SECRET_KEY"] = secret_key
    os.environ["JAWANNDENN_SQLITE_FILE"] = os.path.expanduser(options.database_sqlite3)
    os.environ["JAWANNDENN_URL_PREFIX"] = options.url_prefix

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "jawanndenn.settings")

    # Heavy imports are down here to keep --help fast
    from django.core.management import execute_from_command_line

    with patch("sys.stdout", sys.stderr):
        execute_from_command_line(["./manage.py", "migrate"])
        print()

    if options.dumpdata:
        execute_from_command_line(["./manage.py", "dumpdata"])
    elif options.loaddata:
        print("Importing JSON dump -- this may take a few seconds...", file=sys.stderr)
        execute_from_command_line(["./manage.py", "loaddata", os.path.abspath(options.loaddata)])
    else:
        sys.exit(
            subprocess.call(
                [
                    "gunicorn",
                    "--name=jawanndenn",
                    f"--bind={options.host}:{options.port}",
                    "--workers=1",  # due to use of sqlite3 storage
                    "--access-logfile=-",
                    "--logger-class=gunicorn_color.Logger",
                    "jawanndenn.wsgi",
                ]
            )
        )


def main():
    try:
        _inner_main()
    except KeyboardInterrupt:
        sys.exit(128 + signal.SIGINT)


if __name__ == "__main__":
    main()



# Copyright (C) 2023 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

# These are to allow annotated use of .assertNumQueries
SAVEPOINT = 1
RELEASE_SAVEPOINT = 1


def SELECT(comment):
    return 1



# Generated by Django 2.2.7 on 2019-12-01 16:39

import django.db.models.deletion
import django_extensions.db.fields
from django.db import migrations, models

import jawanndenn.models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Poll",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                (
                    "slug",
                    models.CharField(
                        default=jawanndenn.models._get_random_sha256, max_length=64, unique=True
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("equal_width", models.BooleanField(default=False)),
            ],
            options={
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="PollOption",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("position", models.PositiveSmallIntegerField()),
                ("name", models.CharField(max_length=255)),
                (
                    "poll",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="options",
                        to="jawanndenn.Poll",
                    ),
                ),
            ],
            options={
                "unique_together": {("poll", "position")},
            },
        ),
        migrations.CreateModel(
            name="Ballot",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                ("voter_name", models.CharField(max_length=255)),
                (
                    "poll",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ballots",
                        to="jawanndenn.Poll",
                    ),
                ),
            ],
            options={
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Vote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("yes", models.BooleanField()),
                (
                    "ballot",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="votes",
                        to="jawanndenn.Ballot",
                    ),
                ),
                (
                    "option",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="jawanndenn.PollOption"
                    ),
                ),
            ],
            options={
                "unique_together": {("ballot", "option")},
            },
        ),
    ]



# Copyright (C) 2020 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from django.core.management import BaseCommand

from ...models import Poll


class Command(BaseCommand):
    help = "Deletes expired polls"

    def handle(self, *args, **options):
        polls_deleted, _ = Poll.objects.expired().delete()
        self.stdout.write(self.style.SUCCESS(f"{polls_deleted} poll(s) deleted"))



# Copyright (c) 2024 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from django.http import FileResponse
from django.test import TestCase
from django.urls import reverse_lazy
from parameterized import parameterized


class SecurityTxtTest(TestCase):
    @parameterized.expand(
        [
            (
                "default location",
                "/.well-known/security.txt",
            ),
            (
                "legacy location",
                "/security.txt",
            ),
            (
                "through reverse",
                reverse_lazy("security_txt", kwargs={"path": "security.txt"}),
            ),
        ]
    )
    def test_file_served_properly(self, _label, url):
        self.assertTrue(url.endswith("/security.txt"))

        response = self.client.get(url)

        self.assertIsInstance(response, FileResponse)
        self.assertEqual(response.headers["Content-Type"], "text/plain")



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

_REPLACEMENTS_IN_ORDER = (
    ("**", "<strong>", "</strong>"),
    ("*", "<em>", "</em>"),
    ("__", "<strong>", "</strong>"),
    ("_", "<em>", "</em>"),
    ("`", "<tt>", "</tt>"),
)

_CLOSING_OF = {prefix: closing for prefix, _, closing in _REPLACEMENTS_IN_ORDER}


def safe_html(text):
    if not isinstance(text, str):
        raise ValueError("Not a string: %s" % text)

    # KEEP IN SYNC with javascript client side!

    text = text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    chunks = []

    opened = []
    while text:
        for prefix, opening, closing in _REPLACEMENTS_IN_ORDER:
            if text.startswith(prefix):
                if opened and opened[-1] == prefix:
                    # Close tag
                    chunks.append(closing)
                    opened.pop()
                else:
                    # Open tag
                    chunks.append(opening)
                    opened.append(prefix)

                text = text[len(prefix) :]
                break
        else:
            chunks.append(text[0])
            text = text[1:]

    # Close all unclosed tags
    for prefix in reversed(opened):
        chunks.append(_CLOSING_OF[prefix])

    return "".join(chunks)



# Copyright (C) 2016-2021 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

APP_NAME = "jawanndenn"

_VERSION = (2, 7, 0)
VERSION_STR = ".".join(map(str, _VERSION))



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from factory import Sequence, SubFactory
from factory.django import DjangoModelFactory

from jawanndenn.models import Ballot, Poll, PollOption, Vote


class PollFactory(DjangoModelFactory):
    class Meta:
        model = Poll

    title = Sequence(lambda n: f"Title {n}")


class PollOptionFactory(DjangoModelFactory):
    class Meta:
        model = PollOption

    poll = SubFactory(PollFactory)
    position = Sequence(int)
    name = Sequence(lambda n: f"Option {n}")


class BallotFactory(DjangoModelFactory):
    class Meta:
        model = Ballot

    poll = SubFactory(PollFactory)
    voter_name = Sequence(lambda n: f"User {n}")


class VoteFactory(DjangoModelFactory):
    class Meta:
        model = Vote

    option = SubFactory(PollOptionFactory)



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from unittest import TestCase

from jawanndenn.markup import safe_html


class SafeHtmlTest(TestCase):
    def test_vanilla(self):
        self.assertEqual(safe_html("text"), "text")

    def test_html(self):
        self.assertEqual(safe_html("<b>&nbsp;</b>"), "&lt;b&gt;&amp;nbsp;&lt;/b&gt;")

    def test_bold(self):
        self.assertEqual(safe_html("**text**"), "<strong>text</strong>")
        self.assertEqual(safe_html("__text__"), "<strong>text</strong>")

    def test_italic(self):
        self.assertEqual(safe_html("*text*"), "<em>text</em>")
        self.assertEqual(safe_html("_text_"), "<em>text</em>")

    def test_monospace(self):
        self.assertEqual(safe_html("`text`"), "<tt>text</tt>")

    def test_combined(self):
        self.assertEqual(safe_html("`__*text*__`"), "<tt><strong><em>text</em></strong></tt>")
        self.assertEqual(safe_html("`**_text_**`"), "<tt><strong><em>text</em></strong></tt>")
        self.assertEqual(safe_html("_*text*_"), "<em><em>text</em></em>")

    def test_bad_nesting(self):
        self.assertEqual(
            safe_html("*__text*__"), "<em><strong>text<em><strong></strong></em></strong></em>"
        )

    def test_non_string(self):
        with self.assertRaises(ValueError):
            safe_html(123)



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from functools import wraps


class _XForwardedForHeaderAbsentException(ValueError):
    pass


def _extract_ip_from_x_forwarded_for_header(request):
    """
    Extract IP address for future use with REMOTE_ADDR header
    from header HTTP_X_FORWARDED_FOR.
    May raise exception _XForwardedForHeaderAbsentException .
    """
    try:
        value = request.META["HTTP_X_FORWARDED_FOR"]
    except KeyError:
        raise _XForwardedForHeaderAbsentException

    # NOTE: This assumes that the outermost trusted
    #       reverse proxy is resetting header X-Forwarded-For
    #       rather than appending to an attacker controlled
    #       value from the client request headers.  Else
    #       we would need to use the <n>-rightmost value
    #       instead, where <n> is the number of trusted
    #       reverse proxies in front of the application.
    return value.split(",")[0].strip()


def set_remote_addr_to_x_forwarded_for(get_response):
    """
    Allow use of rate limiting key "ip" and "user_or_ip"
    (that looks at header REMOTE_ADDR) from behind a reverse proxy.

    For more details:
    https://django-ratelimit.readthedocs.io/en/latest/security.html#middleware
    """

    @wraps(get_response)
    def process_request(request):
        try:
            request.META["REMOTE_ADDR"] = _extract_ip_from_x_forwarded_for_header(request)
        except _XForwardedForHeaderAbsentException:
            pass

        return get_response(request)

    return process_request



#! /usr/bin/env python3
# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import os

from setuptools import find_packages, setup

from jawanndenn.metadata import APP_NAME, VERSION_STR


def _read(filename):
    with open(filename) as f:
        return f.read()


_tests_require = [
    "factory-boy>=2.12.0",
    "parameterized>=0.7.1",
]

_extras_require = {
    "tests": _tests_require,
}


def _collect_package_data(top_directory):
    for root, dirs, files in os.walk(os.path.join(top_directory, "static")):
        if files:
            relative_root = os.path.relpath(root, top_directory)
            yield os.path.join(relative_root, "*")


if __name__ == "__main__":
    setup(
        name=APP_NAME,
        version=VERSION_STR,
        license="AGPLv3+",
        description="Libre alternative to Doodle",
        long_description=_read("README.md"),
        long_description_content_type="text/markdown",
        author="Sebastian Pipping",
        author_email="sebastian@pipping.org",
        url="https://github.com/hartwork/jawanndenn",
        python_requires=">=3.9",
        setup_requires=[
            "setuptools>=38.6.0",  # for long_description_content_type
        ],
        install_requires=[
            "django>=2.2.7",
            "django-extensions>=2.2.5",
            "django-ratelimit>=4.0.0",
            "djangorestframework>=3.11.0",
            "gunicorn>=20.0.4",
            "gunicorn-color>=0.1.0",
            "python-dateutil>=2.8.1",
            "python-rapidjson>=1.0",
            "setuptools<81",  # for pkg_resources
        ],
        extras_require=_extras_require,
        tests_require=_tests_require,
        packages=find_packages(),
        package_data={
            APP_NAME: list(_collect_package_data(APP_NAME)),
        },
        entry_points={
            "console_scripts": [
                f"{APP_NAME} = {APP_NAME}.__main__:main",
            ],
        },
        classifiers=[
            "Development Status :: 4 - Beta",
            "Framework :: Django",
            "Framework :: Django :: 2.2",
            "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",  # noqa: E501
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",  # noqa: E501
            "Programming Language :: JavaScript",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Programming Language :: Python :: 3 :: Only",
            "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
            "Topic :: Office/Business :: Scheduling",
        ],
    )



# Copyright (c) 2024 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import pkg_resources
from django.urls import re_path
from django.views.static import serve


def security_txt_urlpatterns(name="security_txt"):
    return [
        re_path(
            "^(?:\\.well-known/)?(?P<path>security\\.txt)$",
            serve,
            kwargs={
                "document_root": pkg_resources.resource_filename(
                    "jawanndenn", "static/well-known"
                ),
            },
            name=name,
        ),
    ]



# Copyright (C) 2020 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from datetime import timedelta
from io import StringIO

from django.test import TestCase
from django.utils.timezone import now

from ....models import Poll
from ....tests.factories import PollFactory
from ..prune_expired_polls import Command


class PruneExpiredPollsTest(TestCase):
    def test_deletes_expired_polls_and_those_only(self):
        expired_poll = PollFactory(expires_at=now())
        not_yet_expired_poll = PollFactory(expires_at=now() + timedelta(minutes=1))
        never_expiring_poll = PollFactory(expires_at=None)

        stdout = StringIO()
        Command(stdout=stdout).handle()

        with self.assertRaises(Poll.DoesNotExist):
            expired_poll.refresh_from_db()

        for poll in (not_yet_expired_poll, never_expiring_poll):
            poll.refresh_from_db()  # would raise DoesNotExist if deleted

        self.assertIn("1 poll(s) deleted", stdout.getvalue())



# Copyright (C) 2019 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

from functools import wraps

import rapidjson as json  # lgtm [py/import-and-import-from]
from django.conf import settings
from django.db import transaction
from django.http import HttpResponseBadRequest, HttpResponseNotFound, JsonResponse
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.views.decorators.http import require_GET, require_POST
from django.views.defaults import bad_request
from rapidjson import JSONDecodeError
from rest_framework.exceptions import ValidationError

from jawanndenn import DEFAULT_MAX_POLLS, DEFAULT_MAX_VOTES_PER_POLL
from jawanndenn.markup import safe_html
from jawanndenn.models import Ballot, Poll, Vote
from jawanndenn.serializers import PollConfigSerializer


def _except_poll_does_not_exist(wrappee):
    """Decorator that turns Poll.DoesNotExist into 404 Not Found"""

    @wraps(wrappee)
    def wrapper(*args, **kwargs):
        try:
            return wrappee(*args, **kwargs)
        except Poll.DoesNotExist:
            return HttpResponseNotFound("No such poll")

    return wrapper


def _except_validation_error(wrappee):
    """Decorator that turns ValidationError into 400 Bad Request"""

    @wraps(wrappee)
    def wrapper(request, *args, **kwargs):
        try:
            return wrappee(request, *args, **kwargs)
        except ValidationError as exception:
            return bad_request(request, exception)

    return wrapper


@require_GET
def index_get_view(request):
    return TemplateResponse(request, template="html/setup.htm")


@require_POST
@_except_validation_error
def poll_post_view(request):
    config_json = request.POST.get("config", "{}")
    try:
        config = json.loads(config_json)
    except JSONDecodeError:
        raise ValidationError("Poll configuration is not well-formed JSON.")

    serializer = PollConfigSerializer(data=config)
    serializer.is_valid(raise_exception=True)

    with transaction.atomic():
        max_polls = getattr(settings, "JAWANNDENN_MAX_POLLS", DEFAULT_MAX_POLLS)
        if Poll.objects.count() >= max_polls:
            return HttpResponseBadRequest(
                f"Maximum number of {max_polls} polls reached, please contact the administrator."
            )

        poll = serializer.save()

    return redirect(poll)


@require_GET
@_except_poll_does_not_exist
def poll_data_get_view(request, poll_id):
    with transaction.atomic():
        poll = Poll.objects.get(slug=poll_id)
        poll_config = {
            "equal_width": poll.equal_width,
            "title": poll.title,
            "options": list(poll.options.order_by("position").values_list("name", flat=True)),
        }

        votes = []
        ballot = None
        ballot_votes = []

        for vote in (
            Vote.objects.filter(ballot__poll=poll)
            .select_related("ballot")
            .order_by("ballot__created", "ballot__id", "option__position")
        ):
            if vote.ballot != ballot:
                if ballot is not None:
                    votes.append([ballot.voter_name, ballot_votes])
                ballot = vote.ballot
                ballot_votes = []
            ballot_votes.append(vote.yes)

        if ballot is not None:
            votes.append([ballot.voter_name, ballot_votes])

    data = {
        "config": poll_config,
        "votes": votes,
    }

    return JsonResponse(data)


@require_GET
@_except_poll_does_not_exist
def poll_get_view(request, poll_id):
    Poll.objects.get(slug=poll_id)

    return TemplateResponse(request, template="html/poll.htm")


@require_POST
@_except_poll_does_not_exist
def vote_post_view(request, poll_id):
    with transaction.atomic():
        poll = Poll.objects.get(slug=poll_id)

        max_votes_per_poll = getattr(
            settings, "JAWANNDENN_MAX_VOTES_PER_POLL", DEFAULT_MAX_VOTES_PER_POLL
        )
        if poll.ballots.count() >= max_votes_per_poll:
            return HttpResponseBadRequest(
                f"Maximum number of {max_votes_per_poll} "
                "votes reached for this poll"
                ", please contact the administrator."
            )

        voter_name = safe_html(request.POST.get("voterName"))
        votes = [
            request.POST.get(f"option{i}", "off") == "on" for i in range(poll.options.count())
        ]

        ballot = Ballot.objects.create(poll=poll, voter_name=voter_name)
        for option, vote in zip(poll.options.order_by("position"), votes):
            Vote.objects.create(ballot=ballot, option=option, yes=vote)

    return redirect(poll)
